r"""
# pycitysim

City Simulator and OpenCity databases Python SDK

## 安装 Installation

```shell
pip install pycitysim
```

安装 grpcio 的过程中，如果出现：
When installing grpcio, you may encounter the following error:

```
pip install fails with "No such file or directory: 'c++': 'c++'"
```

代表缺少 C++相关依赖。在 Debian 镜像上，执行：
It means that C++ related dependencies are missing. On the Debian image, execute:

```shell
sudo apt install build-essential
```

在 alpine 镜像上，执行：
On the alpine image, execute:

```shell
apk add g++
```

## 主要子库 Sub-packages

- pycitysim.apphub: OpenCity后端交互库。OpenCity backend interaction package.
- pycitysim.map: 地图数据操作库。Map data operation package.
- pycitysim.routing: 路径规划操作库。Routing operation package.
- pycitysim.sim: 模拟器gRPC接入客户端。City Simulator gRPC access client.
- pycitysim.urbankg: 城市知识图谱访问模块。City knowledge graph access module.

## 示例代码

访问 [examples](https://github.com/tsinghua-fib-lab/pycitysim/-/tree/main/examples?ref_type=heads) 查看示例代码。

"""

from . import map, routing, sidecar, sim, urbankg, utils, apphub

__all__ = [
    "map",
    "routing",
    "sidecar",
    "sim",
    "urbankg",
    "utils",
    "apphub",
]



from typing import Any, Awaitable, Coroutine, cast, Union, Dict

import grpc
from google.protobuf.json_format import ParseDict
from pycityproto.city.map.v2 import lane_service_pb2 as lane_service
from pycityproto.city.map.v2 import lane_service_pb2_grpc as lane_grpc

from ..utils.protobuf import async_parse

__all__ = ["LaneService"]


class LaneService:
    """
    交通模拟lane服务
    Traffic simulation lane service
    """

    def __init__(self, aio_channel: grpc.aio.Channel):
        self._aio_stub = lane_grpc.LaneServiceStub(aio_channel)

    def GetLane(
        self, req: Union[lane_service.GetLaneRequest, dict], dict_return: bool = True
    ) -> Coroutine[Any, Any, Union[Dict[str, Any], lane_service.GetLaneResponse]]:
        """
        获取Lane的信息
        Get Lane's information

        Args:
        - req (dict): https://cityproto.sim.fiblab.net/#city.map.v2.GetLaneRequest

        Returns:
        - https://cityproto.sim.fiblab.net/#city.map.v2.GetLaneResponse
        """
        if type(req) != lane_service.GetLaneRequest:
            req = ParseDict(req, lane_service.GetLaneRequest())
        res = cast(Awaitable[lane_service.GetLaneResponse], self._aio_stub.GetLane(req))
        return async_parse(res, dict_return)

    def SetLaneMaxV(
        self,
        req: Union[lane_service.SetLaneMaxVRequest, dict],
        dict_return: bool = True,
    ) -> Coroutine[Any, Any, Union[Dict[str, Any], lane_service.SetLaneMaxVResponse]]:
        """
        设置Lane的最大速度（限速）
        Set the maximum speed of Lane (speed limit)

        Args:
        - req (dict): https://cityproto.sim.fiblab.net/#city.map.v2.SetLaneMaxVRequest

        Returns:
        - https://cityproto.sim.fiblab.net/#city.map.v2.SetLaneMaxVResponse
        """
        if type(req) != lane_service.SetLaneMaxVRequest:
            req = ParseDict(req, lane_service.SetLaneMaxVRequest())
        res = cast(
            Awaitable[lane_service.SetLaneMaxVResponse], self._aio_stub.SetLaneMaxV(req)
        )
        return async_parse(res, dict_return)

    def GetLaneByLongLatBBox(
        self,
        req: Union[lane_service.GetLaneByLongLatBBoxRequest, dict],
        dict_return: bool = True,
    ) -> Coroutine[
        Any, Any, Union[Dict[str, Any], lane_service.GetLaneByLongLatBBoxResponse]
    ]:
        """
        获取特定区域内的Lane的信息
        Get lane information in a specific region

        Args:
        - req (dict): https://cityproto.sim.fiblab.net/#city.map.v2.GetLaneByLongLatBBoxRequest

        Returns:
        - https://cityproto.sim.fiblab.net/#city.map.v2.GetLaneByLongLatBBoxResponse
        """
        if type(req) != lane_service.GetLaneByLongLatBBoxRequest:
            req = ParseDict(req, lane_service.GetLaneByLongLatBBoxRequest())
        res = cast(
            Awaitable[lane_service.GetLaneByLongLatBBoxResponse],
            self._aio_stub.GetLaneByLongLatBBox(req),
        )
        return async_parse(res, dict_return)



import grpc

__all__ = ["create_aio_channel"]


def create_aio_channel(server_address: str, secure: bool = False) -> grpc.aio.Channel:
    """
    创建一个grpc的异步channel
    Create a grpc asynchronous channel

    Args:
    - server_address (str): 服务器地址。server address.
    - secure (bool, optional): 是否使用安全连接. Defaults to False. Whether to use a secure connection. Defaults to False.

    Returns:
    - grpc.aio.Channel: grpc的异步channel。grpc asynchronous channel.
    """
    if server_address.startswith("http://"):
        server_address = server_address.split("//")[1]
        if secure:
            raise ValueError("secure channel must use `https` or not use `http`")
    elif server_address.startswith("https://"):
        server_address = server_address.split("//")[1]
        if not secure:
            secure = True

    if secure:
        return grpc.aio.secure_channel(server_address, grpc.ssl_channel_credentials())
    else:
        return grpc.aio.insecure_channel(server_address)



from typing import Any, Awaitable, Coroutine, cast, Union, Dict

import grpc
from google.protobuf.json_format import ParseDict
from pycityproto.city.social.v1 import social_service_pb2 as social_service
from pycityproto.city.social.v1 import social_service_pb2_grpc as social_grpc

from ..utils.protobuf import async_parse

__all__ = ["SocialService"]


class SocialService:
    """
    城市模拟社交服务
    City simulation social service
    """

    def __init__(self, aio_channel: grpc.aio.Channel):
        self._aio_stub = social_grpc.SocialServiceStub(aio_channel)

    def Send(
        self, req: Union[social_service.SendRequest, dict], dict_return: bool = True
    ) -> Coroutine[Any, Any, Union[Dict[str, Any], social_service.SendResponse]]:
        """
        发送消息
        Send message

        Args:
        - req (dict): https://cityproto.sim.fiblab.net/#city.social.v1.SendRequest

        Returns:
        - https://cityproto.sim.fiblab.net/#city.social.v1.SendResponse
        """
        if type(req) != social_service.SendRequest:
            req = ParseDict(req, social_service.SendRequest())
        res = cast(Awaitable[social_service.SendResponse], self._aio_stub.Send(req))
        return async_parse(res, dict_return)

    def Receive(
        self, req: Union[social_service.ReceiveRequest, dict], dict_return: bool = True
    ) -> Coroutine[Any, Any, Union[Dict[str, Any], social_service.ReceiveResponse]]:
        """
        接收消息
        Receive message
        
        Args:
        - req (dict): https://cityproto.sim.fiblab.net/#city.social.v1.ReceiveRequest

        Returns:
        - https://cityproto.sim.fiblab.net/#city.social.v1.ReceiveResponse
        """
        if type(req) != social_service.ReceiveRequest:
            req = ParseDict(req, social_service.ReceiveRequest())
        res = cast(
            Awaitable[social_service.ReceiveResponse], self._aio_stub.Receive(req)
        )
        return async_parse(res, dict_return)



"""
模拟器gRPC接入客户端
Emulator gRPC access client
"""

from .person_service import PersonService
from .aoi_service import AoiService
from .client import CityClient
from .clock_service import ClockService
from .economy_services import EconomyOrgService, EconomyPersonService
from .event_service import EventService
from .lane_service import LaneService
from .road_service import RoadService
from .social_service import SocialService

__all__ = [
    "CityClient",
    "ClockService",
    "PersonService",
    "AoiService",
    "LaneService",
    "RoadService",
    "SocialService",
    "EconomyPersonService",
    "EconomyOrgService",
    "EventService",
]



import base64
from dataclasses import dataclass
from io import BytesIO
from typing import Any, Dict, Optional, List

from PIL.Image import Image
import geojson
import requests
from requests.auth import HTTPBasicAuth

__all__ = ["AppHubClient", "AgentMessage", "UserMessage"]


def _image2base64(image: Image) -> str:
    buffered = BytesIO()
    image.save(buffered, format="PNG")
    return base64.b64encode(buffered.getvalue()).decode("utf-8")


@dataclass
class AgentMessage:
    id: int
    """
    发送者Agent ID
    Agent ID of sender
    """
    timestamp: int
    """
    消息时间戳（单位：毫秒）
    Message timestamp (milliseconds)
    """
    content: str
    """
    消息内容
    Message content
    """
    images: Optional[List[Image]]
    """
    消息图片
    Message images
    """
    sub_messages: Optional[List["AgentMessage"]]

    def to_dict(self):
        d = {
            "id": self.id,
            "timestamp": self.timestamp,
            "content": self.content,
        }
        if self.images is not None:
            d["images"] = [_image2base64(image) for image in self.images]
        if self.sub_messages is not None:
            d["subMessages"] = [message.to_dict() for message in self.sub_messages]
        return d


@dataclass
class UserMessage:
    id: int
    """
    发送者User ID对应的Agent ID
    Agent ID corresponding to the sender User ID
    """
    timestamp: int
    """
    消息时间戳（单位：毫秒）
    Message timestamp (milliseconds)
    """
    content: str
    """
    消息内容
    Message content
    """


class AppHubClient:
    """
    AppHub客户端
    AppHub client

    # 操作流程
    # Operating procedures

    1. 初始化AppHubClient，填入从前端获得的app_id和app_secret
    1. Initialize AppHubClient, fill in the app_id and app_secret obtained from the front end    
    2. 调用bind_person/bind_org绑定模拟器内的人，获得agent_id
    2. Call bind_person/bind_org to bind the person in the simulator, and obtain the agent_id
    3. 调用update_agent_map更新agent的地图内容，以改变agent在地图上的显示
    3. Call update_agent_map to update the agent's map content, to change the agent's display on the map
    4. 调用update_agent_messages更新agent的消息内容，以改变agent的消息内容
    4. Call update_agent_messages to update the agent's message content, to change the agent's message content
    5. 定期调用fetch_user_message获取用户发送给agent的消息，并进行处理
    5. Call fetch_user_message regularly to obtain the messages sent by the user to the agent, and process them
    6. 如果不再使用agent，调用release_agent释放agent，以允许其他app绑定
    6. If the agent is no longer used, call release_agent to release the agent to allow other apps to bind.
    """

    def __init__(
        self,
        app_hub_url: str,
        app_id: int,
        app_secret: str,
        proxies: Optional[dict] = None,
    ):
        """
        Args:
        - app_id: app ID，从前端注册页面获取。app ID, obtained from the frontend registration page.
        - app_secret: app secret，从前端注册页面获取。app secret, obtained from the frontend registration page.
        """
        self.app_hub_url = app_hub_url
        self.app_id = app_id
        self.app_secret = app_secret
        self.proxies = proxies
        self.agents = {}  # agent_id -> agent (bind body)

        self._auth = HTTPBasicAuth(str(app_id), app_secret)

    def _bind(self, foreign_id: int, type: str, name: str, avatar: Image) -> int:
        body = {
            "type": type,
            "name": name,
            "avatar": _image2base64(avatar),
            "foreignID": foreign_id,
        }
        res = requests.post(
            self.app_hub_url + "/agents",
            json=body,
            auth=self._auth,
            proxies=self.proxies,
        )
        if res.status_code != 200:
            raise Exception(f"[{res.status_code}] AppHub bind failed: " + res.text)
        data = res.json()
        agent_id = data["data"]["id"]
        self.agents[agent_id] = body
        return agent_id

    def bind_person(self, person_id: int, name: str, avatar: Image):
        """
        绑定模拟器内的人，只有绑定的person才能被访问
        Bind person in the simulator. Only bound person can be accessed

        Args:
        - person_id: 模拟器内的人的ID。The ID of the person in the simulator.
        - name: 人的名字（前端显示）。Person's name (displayed on the front end).
        - avatar: 人的头像（前端显示）。Person's avatar (displayed on the front end).

        Returns:
        - agent_id: 绑定后的agent ID。the bound agent ID.

        Raises:
        - Exception: 绑定失败。Binding failed.
        """
        return self._bind(person_id, "person", name, avatar)

    def bind_org(self, org_id: int, name: str, avatar: Image):
        """
        绑定模拟器内的组织，只有绑定的org才能被访问
        Bind the organization in the simulator. Only the bound org can be accessed

        Args:
        - org_id: 模拟器内的组织的ID。ID of the organization in the simulator
        - name: 组织的名字（前端显示）。Organization's name (displayed on the front end).
        - avatar: 组织的头像（前端显示）。Organization's avatar (displayed on the front end).

        Returns:
        - agent_id: 绑定后的agent ID。the bound agent ID.

        Raises:
        - Exception: 绑定失败。Binding failed.
        """
        return self._bind(org_id, "org", name, avatar)

    def release_agent(self, agent_id: int) -> bool:
        """
        释放agent(person/org)，释放后agent将不再被访问并允许其他app绑定
        Release the agent (person/org). After the release, the agent will no longer be accessed and allows other apps to bind.

        Args:
        - agent_id: agent的ID。ID of the agent.

        Returns:
        - bool: 是否成功。whether succeed.
        """
        res = requests.delete(
            self.app_hub_url + "/agents/" + str(agent_id),
            auth=self._auth,
            proxies=self.proxies,
        )
        return res.status_code == 200

    def update_agent_map(
        self,
        agent_id: int,
        lnglat: List[float],
        geojsons: Optional[geojson.FeatureCollection] = None,
        street_view: Optional[Image] = None,
        popup: Optional[str] = None,
    ):
        """
        更新agent的地图内容
        Update the agent’s map content

        Args:
        - agent_id: agent的ID。ID of the agent.
        - lnglat: 经纬度，格式为 [lng, lat]。Latitude and longitude in [lng, lat].
        - geojsons: geojson FeatureCollection，用于在地图上显示点线面。geojson FeatureCollection, used to display points, lines and polygons on the map.
        - street_view: 街景图片。Street view images.
        - popup: 弹出框内容。Pop-up box content.

        Raises:
        - Exception: 更新失败。Updating failed.
        """

        body: Dict[str, Any] = {
            "lnglat": lnglat,
        }
        if geojsons is not None:
            body["geojsons"] = geojsons
        if street_view is not None:
            body["streetView"] = _image2base64(street_view)
        if popup is not None:
            body["popup"] = popup

        res = requests.put(
            self.app_hub_url + "/agents/" + str(agent_id) + "/map",
            json=body,
            auth=self._auth,
            proxies=self.proxies,
        )
        if res.status_code != 200:
            raise Exception(
                f"[{res.status_code}] AppHub update map failed: " + res.text
            )

    def update_agent_messages(self, agent_id: int, messages: List[AgentMessage]):
        """
        更新agent的消息
        Update agent's messages

        Args:
        - agent_id: agent的ID。ID of the agent.
        - messages: 消息列表（注意顺序，最新的消息在最后）。Message list (note the order, the latest message at the end).

        Raises:
        - Exception: 更新失败。Updating failed.
        """
        res = requests.put(
            self.app_hub_url + "/agents/" + str(agent_id) + "/messages",
            json={"messages": [message.to_dict() for message in messages]},
            auth=self._auth,
            proxies=self.proxies,
        )
        if res.status_code != 200:
            raise Exception(
                f"[{res.status_code}] AppHub update messages failed: " + res.text
            )

    def fetch_user_messages(
        self,
        agent_id: int,
    ) -> List[UserMessage]:
        """
        获取用户发送给agent的消息
        Fetch the message sent by the user to the agent

        Args:
        - agent_id: agent的ID。ID of the agent.

        Returns:
        - messages: 消息列表。Message list.

        Raises:
        - Exception: 获取失败。Fetching failed.
        """
        res = requests.post(
            self.app_hub_url + "/agents/" + str(agent_id) + "/fetch-user-messages",
            auth=self._auth,
            proxies=self.proxies,
        )
        if res.status_code != 200:
            raise Exception(
                f"[{res.status_code}] AppHub fetch messages failed: " + res.text
            )
        data = res.json()
        return [UserMessage(**message) for message in data["data"]]



from typing import List, Optional, Tuple

from lru import LRU
from pymongo import MongoClient

__all__ = ["UrbanKG"]


class UrbanKG:
    """
    城市知识图谱
    Urban knowledge graph
    """

    CLASSES = ["Region", "POI", "Cate3", "Brand", "Ba", "Cate2", "Cate1"]
    RELATIONS = [
        "SubCateOf_3to2",
        "RelatedBrand",
        "NearBy",
        "LocateAt",
        "LargeOD",
        "Competitive_zhishi",
        "CoCheckin",
        "BorderBy",
        "SubCateOf_2to1",
        "Brand2Cate1",
        "Brand2Cate2",
        "SubCateOf_3to1",
        "BaServe",
        "SimilarPOIs",
        "Cate3Of",
        "Brand2Cate3",
        "BrandOf",
        "Cate2Of",
        "BelongTo",
        "Cate1Of",
    ]

    def __init__(
        self,
        mongo_uri: str,
        mongo_db: str,
        mongo_entity_coll: str,
        mongo_relation_coll: str,
        cache_size: int = 100,
    ):
        """
        Args:
        - mongo_uri: MongoDB URI
        - mongo_db: MongoDB数据库名。MongoDB database name.
        - mongo_entity_coll: MongoDB实体集合名。MongoDB entity collection name.
        - mongo_relation_coll: MongoDB关系集合名。MongoDB relation collection name.
        - cache_size: LRU缓存大小。LRU cache size.
        """
        client = MongoClient(mongo_uri)
        db = client[mongo_db]
        self._entity_coll_name = mongo_entity_coll
        self._entity_coll = db[mongo_entity_coll]
        self._relation_coll_name = mongo_relation_coll
        self._relation_coll = db[mongo_relation_coll]

        self._entity_cache = LRU(cache_size)  # query key => result
        self._relation_cache = LRU(cache_size)  # query key => result

    def get_entity(self, id: str) -> Optional[dict]:
        """
        查询实体
        query entity

        Args:
        - id: 实体ID。entity ID.

        Returns:
        - Optional[dict]: 实体，为None表示不存在。Entity, None means it does not exist.
        """
        if id in self._entity_cache:
            return self._entity_cache[id]
        result = self._entity_coll.find_one({"id": id})
        if result is not None:
            del result["_id"]
        self._entity_cache[id] = result
        return result

    def query_object(self, subject_id: str, relation: str) -> List[str]:
        """
        根据主体（subject）和关系（relation）查询客体（object）
        Query objects based on subject and relationship

        Args:
        - subject_id: 主体ID。subject ID.
        - relation: 关系。relation.

        Returns
        - List[str]: 客体ID列表
        """
        key = (subject_id, relation, "")
        if key in self._relation_cache:
            return self._relation_cache[key]
        result = list(
            self._relation_coll.aggregate(
                [
                    {
                        "$match": {
                            "subject": subject_id,
                            "relation": relation,
                        }
                    },
                    {
                        "$lookup": {
                            "from": self._entity_coll_name,
                            "localField": "object",
                            "foreignField": "id",
                            "as": "object",
                        }
                    },
                    {"$unwind": {"path": "$object"}},
                    {"$project": {"object": 1}},
                ]
            )
        )
        objects = [r["object"] for r in result]
        for obj in objects:
            del obj["_id"]
        self._relation_cache[key] = objects
        return objects

    def query_subject(self, object_id: str, relation: str) -> List[str]:
        """
        根据客体（object）和关系（relation）查询主体（subject）
        Query the subject based on the object and relationship

        Args:
        - object_id: 客体ID。object ID.
        - relation: 关系。relation.

        Returns
        - List[str]: 主体ID列表。List of subject IDs.
        """
        key = ("", relation, object_id)
        if key in self._relation_cache:
            return self._relation_cache[key]
        result = list(
            self._relation_coll.aggregate(
                [
                    {
                        "$match": {
                            "relation": relation,
                            "object": object_id,
                        }
                    },
                    {
                        "$lookup": {
                            "from": self._entity_coll_name,
                            "localField": "subject",
                            "foreignField": "id",
                            "as": "subject",
                        }
                    },
                    {"$unwind": {"path": "$subject"}},
                    {"$project": {"subject": 1}},
                ]
            )
        )
        subjects = [r["subject"] for r in result]
        for sub in subjects:
            del sub["_id"]
        self._relation_cache[key] = subjects
        return subjects

    def explore(self, subject_id: str, relations: List[str]) -> List[Tuple[str, dict]]:
        """
        根据主体（subject）和要探索的关系列表（relations）查询所有有关的客体（object）
        Query all related objects based on the subject and the list of relationships to be explored.

        Args:
        - subject_id: 主体ID。subject ID.
        - relations: 要探索的关系列表。The list of relationships to be explored.

        Returns:
        - List[Tuple[str, dict]]: 结果列表，Tuple第一项为关系，第二项为客体。Result list, the first item of Tuple is the relationship, the second item is the object.
        """
        key = (subject_id, tuple(relations))
        if key in self._relation_cache:
            return self._relation_cache[key]
        result = list(
            self._relation_coll.aggregate(
                [
                    {
                        "$match": {
                            "subject": subject_id,
                            "relation": {"$in": relations},
                        }
                    },
                    {
                        "$lookup": {
                            "from": self._entity_coll_name,
                            "localField": "object",
                            "foreignField": "id",
                            "as": "object",
                        }
                    },
                    {"$unwind": {"path": "$object"}},
                    {"$project": {"relation": 1, "object": 1}},
                ]
            )
        )
        objects = [(r["relation"], r["object"]) for r in result]
        for _, obj in objects:
            del obj["_id"]
        self._relation_cache[key] = objects

        return objects
